"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createModelSelectorSpec = createModelSelectorSpec;
exports.FieldSelectorSpec = exports.MapSelectorSpec = exports.ModelSelectorSpec = exports.SelectorSpec = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _typeof2 = _interopRequireDefault(require("@babel/runtime/helpers/typeof"));

var _objectWithoutProperties2 = _interopRequireDefault(require("@babel/runtime/helpers/objectWithoutProperties"));

var _inheritsLoose2 = _interopRequireDefault(require("@babel/runtime/helpers/inheritsLoose"));

var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _fields = require("./fields");

var _QuerySet = _interopRequireDefault(require("./QuerySet"));

var _Model = _interopRequireDefault(require("./Model"));

/**
 * @module selectors
 */
var ALL_INSTANCES = Symbol('REDUX_ORM_ALL_INSTANCES');

function idArgSelector(state, idArg) {
  return idArg;
}

var SelectorSpec =
/*#__PURE__*/
function () {
  function SelectorSpec(_ref) {
    var parent = _ref.parent,
        orm = _ref.orm;
    this._parent = parent;
    this._orm = orm;
  }

  (0, _createClass2["default"])(SelectorSpec, [{
    key: "cachePath",
    get: function get() {
      var basePath = this._parent ? this._parent.cachePath : [];
      return [].concat((0, _toConsumableArray2["default"])(basePath), [this.key]);
    }
  }]);
  return SelectorSpec;
}();

exports.SelectorSpec = SelectorSpec;

var ModelSelectorSpec =
/*#__PURE__*/
function (_SelectorSpec) {
  (0, _inheritsLoose2["default"])(ModelSelectorSpec, _SelectorSpec);

  function ModelSelectorSpec(_ref2) {
    var _this;

    var model = _ref2.model,
        other = (0, _objectWithoutProperties2["default"])(_ref2, ["model"]);
    _this = _SelectorSpec.call(this, other) || this;
    _this._model = model;
    return _this;
  }

  (0, _createClass2["default"])(ModelSelectorSpec, [{
    key: "key",
    get: function get() {
      return this._model.modelName;
    }
  }, {
    key: "dependencies",
    get: function get() {
      return [this._orm, idArgSelector];
    }
  }, {
    key: "resultFunc",
    get: function get() {
      var _this2 = this;

      return function (_ref3, idArg) {
        var ModelClass = _ref3[_this2._model.modelName];

        if (typeof idArg === 'undefined') {
          return ModelClass.all().toRefArray();
        }

        if (Array.isArray(idArg)) {
          return idArg.map(function (id) {
            var instance = ModelClass.withId(id);
            return instance ? instance.ref : null;
          });
        }

        var instance = ModelClass.withId(idArg);
        return instance ? instance.ref : null;
      };
    }
  }, {
    key: "keySelector",
    get: function get() {
      return function (state, idArg) {
        return typeof idArg === 'undefined' ? ALL_INSTANCES : idArg;
      };
    }
  }]);
  return ModelSelectorSpec;
}(SelectorSpec);

exports.ModelSelectorSpec = ModelSelectorSpec;

var ModelBasedSelectorSpec =
/*#__PURE__*/
function (_SelectorSpec2) {
  (0, _inheritsLoose2["default"])(ModelBasedSelectorSpec, _SelectorSpec2);

  function ModelBasedSelectorSpec(_ref4) {
    var _this3;

    var model = _ref4.model,
        other = (0, _objectWithoutProperties2["default"])(_ref4, ["model"]);
    _this3 = _SelectorSpec2.call(this, other) || this;
    _this3._model = model;
    return _this3;
  }

  (0, _createClass2["default"])(ModelBasedSelectorSpec, [{
    key: "resultFunc",
    get: function get() {
      var _this4 = this;

      return function (session, idArg) {
        for (var _len = arguments.length, other = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
          other[_key - 2] = arguments[_key];
        }

        var ModelClass = session[_this4._model.modelName];

        if (typeof idArg === 'undefined') {
          return ModelClass.all().toModelArray().map(function (instance) {
            return _this4.valueForInstance.apply(_this4, [instance, session].concat(other));
          });
        }

        if (Array.isArray(idArg)) {
          return idArg.map(function (id) {
            return _this4.valueForInstance.apply(_this4, [ModelClass.withId(id), session].concat(other));
          });
        }

        return _this4.valueForInstance.apply(_this4, [ModelClass.withId(idArg), session].concat(other));
      };
    }
  }]);
  return ModelBasedSelectorSpec;
}(SelectorSpec);

var MapSelectorSpec =
/*#__PURE__*/
function (_ModelBasedSelectorSp) {
  (0, _inheritsLoose2["default"])(MapSelectorSpec, _ModelBasedSelectorSp);

  function MapSelectorSpec(_ref5) {
    var _this5;

    var field = _ref5.field,
        accessorName = _ref5.accessorName,
        selector = _ref5.selector,
        other = (0, _objectWithoutProperties2["default"])(_ref5, ["field", "accessorName", "selector"]);
    _this5 = _ModelBasedSelectorSp.call(this, other) || this;
    _this5._field = field;
    _this5._accessorName = accessorName;
    _this5._selector = selector;
    return _this5;
  }

  var _proto = MapSelectorSpec.prototype;

  _proto.valueForInstance = function valueForInstance(instance, session, state) {
    var _this6 = this;

    if (!instance) return null;
    var value = instance[this._accessorName];
    if (!value) return null;
    var mapIdAttribute = session[this._field.toModelName].idAttribute;
    return value.toRefArray().map(function (ref) {
      return _this6._selector(state, ref[mapIdAttribute]);
    });
  };

  (0, _createClass2["default"])(MapSelectorSpec, [{
    key: "key",
    get: function get() {
      return this._selector;
    }
  }, {
    key: "dependencies",
    get: function get() {
      return [this._orm, idArgSelector, function (state) {
        return state;
      }];
    }
  }, {
    key: "keySelector",
    get: function get() {
      return function (state, idArg) {
        return typeof idArg === 'undefined' ? ALL_INSTANCES : idArg;
      };
    }
  }]);
  return MapSelectorSpec;
}(ModelBasedSelectorSpec);

exports.MapSelectorSpec = MapSelectorSpec;

var FieldSelectorSpec =
/*#__PURE__*/
function (_ModelBasedSelectorSp2) {
  (0, _inheritsLoose2["default"])(FieldSelectorSpec, _ModelBasedSelectorSp2);

  function FieldSelectorSpec(_ref6) {
    var _this7;

    var field = _ref6.field,
        fieldModel = _ref6.fieldModel,
        accessorName = _ref6.accessorName,
        isVirtual = _ref6.isVirtual,
        other = (0, _objectWithoutProperties2["default"])(_ref6, ["field", "fieldModel", "accessorName", "isVirtual"]);
    _this7 = _ModelBasedSelectorSp2.call(this, other) || this;
    _this7._field = field;
    _this7._fieldModel = fieldModel;
    _this7._accessorName = accessorName;
    _this7._isVirtual = isVirtual;
    return _this7;
  }

  var _proto2 = FieldSelectorSpec.prototype;

  _proto2.valueForInstance = function valueForInstance(instance, session) {
    if (!instance) return null;
    var value;

    if (this._parent instanceof ModelSelectorSpec) {
      value = instance[this._accessorName];
    } else if (this._parent instanceof FieldSelectorSpec) {
      var ParentToModel = session[this._parent.toModelName];

      var parentRef = this._parent.valueForInstance(instance, session);

      var parentInstance = parentRef ? new ParentToModel(parentRef) : null;
      value = parentInstance ? parentInstance[this._accessorName] : null;
    }

    if (value instanceof _Model["default"]) {
      return value ? value.ref : null;
    }

    if (value instanceof _QuerySet["default"]) {
      return value.toRefArray();
    }

    return value;
  };

  _proto2.map = function map(selector) {
    /* eslint-disable no-underscore-dangle */
    if (selector instanceof ModelSelectorSpec) {
      if (this.toModelName === selector._model.modelName) {
        throw new Error("Cannot select models in a `map()` call. If you just want the `".concat(this._accessorName, "` as a ref array then you can simply drop the `map()`. Otherwise make sure you're passing a field selector of the form `").concat(this.toModelName, ".<field>` or a custom selector instead."));
      } else {
        throw new Error("Cannot select `".concat(selector._model.modelName, "` models in this `map()` call. Make sure you're passing a field selector of the form `").concat(this.toModelName, ".<field>` or a custom selector instead."));
      }
    } else if (selector instanceof FieldSelectorSpec) {
      if (this.toModelName !== selector._model.modelName) {
        throw new Error("Cannot select fields of the `".concat(selector._model.modelName, "` model in this `map()` call. Make sure you're passing a field selector of the form `").concat(this.toModelName, ".<field>` or a custom selector instead."));
      }
    } else if (!selector || typeof selector !== 'function' || !selector.recomputations) {
      throw new Error("`map()` requires a selector as an input. Received: ".concat(JSON.stringify(selector), " of type ").concat((0, _typeof2["default"])(selector)));
    }

    if (!(this._field instanceof _fields.ForeignKey) && !(this._field instanceof _fields.ManyToMany)) {
      throw new Error('Cannot map selectors for non-collection fields');
    }

    return new MapSelectorSpec({
      parent: this,
      model: this._model,
      orm: this._orm,
      field: this._field,
      accessorName: this._accessorName,
      selector: selector
    });
  };

  (0, _createClass2["default"])(FieldSelectorSpec, [{
    key: "key",
    get: function get() {
      return this._accessorName;
    }
  }, {
    key: "dependencies",
    get: function get() {
      return [this._orm, idArgSelector];
    }
  }, {
    key: "keySelector",
    get: function get() {
      return function (state, idArg) {
        return typeof idArg === 'undefined' ? ALL_INSTANCES : idArg;
      };
    }
  }, {
    key: "toModelName",
    get: function get() {
      return this._field.toModelName === 'this' ? this._fieldModel.modelName : this._field.toModelName;
    }
  }]);
  return FieldSelectorSpec;
}(ModelBasedSelectorSpec);

exports.FieldSelectorSpec = FieldSelectorSpec;

function createFieldSelectorSpec(_ref7) {
  var parent = _ref7.parent,
      model = _ref7.model,
      field = _ref7.field,
      accessorName = _ref7.accessorName,
      orm = _ref7.orm,
      isVirtual = _ref7.isVirtual;
  var fieldSelectorSpec = new FieldSelectorSpec({
    parent: parent,
    model: model,
    orm: orm,
    field: field,
    accessorName: accessorName,
    isVirtual: isVirtual
  });
  /* Do not even try to create field selectors below attributes. */

  if (!(field instanceof _fields.RelationalField)) {
    // "orm.Author.name.publisher" would be nonsense
    return fieldSelectorSpec;
  }
  /* Prevent field selectors below collections. */


  if (parent instanceof FieldSelectorSpec) {
    /* eslint-disable no-underscore-dangle */
    if ( // "orm.Author.books.publisher" would be nonsense
    parent._field instanceof _fields.ForeignKey && parent._isVirtual || // "orm.Genre.books.publisher" would be nonsense
    parent._field instanceof _fields.ManyToMany) {
      throw new Error("Cannot create a selector for `".concat(parent._accessorName, ".").concat(accessorName, "` because `").concat(parent._accessorName, "` is a collection field."));
    }
  }

  var toModelName = field.toModelName;
  if (!toModelName) return fieldSelectorSpec;
  var toModel = orm.get(toModelName === 'this' ? model.modelName : toModelName);
  Object.entries(toModel.fields).forEach(function (_ref8) {
    var _ref9 = (0, _slicedToArray2["default"])(_ref8, 2),
        relatedFieldName = _ref9[0],
        relatedField = _ref9[1];

    var fieldAccessorName = relatedField.as || relatedFieldName;
    Object.defineProperty(fieldSelectorSpec, fieldAccessorName, {
      get: function get() {
        return createFieldSelectorSpec({
          parent: fieldSelectorSpec,
          model: model,
          fieldModel: toModel,
          field: relatedField,
          accessorName: fieldAccessorName,
          orm: orm,
          isVirtual: false
        });
      }
    });
  });
  Object.entries(toModel.virtualFields).forEach(function (_ref10) {
    var _ref11 = (0, _slicedToArray2["default"])(_ref10, 2),
        relatedFieldName = _ref11[0],
        relatedField = _ref11[1];

    var fieldAccessorName = relatedField.as || relatedFieldName;

    if (fieldSelectorSpec.hasOwnProperty(fieldAccessorName)) {
      return;
    }

    Object.defineProperty(fieldSelectorSpec, fieldAccessorName, {
      get: function get() {
        return createFieldSelectorSpec({
          parent: fieldSelectorSpec,
          model: model,
          fieldModel: toModel,
          field: relatedField,
          accessorName: fieldAccessorName,
          orm: orm,
          isVirtual: true
        });
      }
    });
  });
  return fieldSelectorSpec;
}

function createModelSelectorSpec(_ref12) {
  var model = _ref12.model,
      orm = _ref12.orm;
  var modelSelectorSpec = new ModelSelectorSpec({
    parent: null,
    orm: orm,
    model: model
  });
  Object.entries(model.fields).forEach(function (_ref13) {
    var _ref14 = (0, _slicedToArray2["default"])(_ref13, 2),
        fieldName = _ref14[0],
        field = _ref14[1];

    var fieldAccessorName = field.as || fieldName;
    Object.defineProperty(modelSelectorSpec, fieldAccessorName, {
      get: function get() {
        return createFieldSelectorSpec({
          parent: modelSelectorSpec,
          model: model,
          fieldModel: model,
          field: field,
          accessorName: fieldAccessorName,
          orm: orm,
          isVirtual: false
        });
      }
    });
  });
  Object.entries(model.virtualFields).forEach(function (_ref15) {
    var _ref16 = (0, _slicedToArray2["default"])(_ref15, 2),
        fieldName = _ref16[0],
        field = _ref16[1];

    var fieldAccessorName = field.as || fieldName;

    if (modelSelectorSpec.hasOwnProperty(fieldAccessorName)) {
      return;
    }

    Object.defineProperty(modelSelectorSpec, fieldAccessorName, {
      get: function get() {
        return createFieldSelectorSpec({
          parent: modelSelectorSpec,
          model: model,
          fieldModel: model,
          field: field,
          accessorName: fieldAccessorName,
          orm: orm,
          isVirtual: true
        });
      }
    });
  });
  return modelSelectorSpec;
}